//-----------------------------------------------------------------------------
// Copyright  1999, 2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// G. M. Adelson-Velskii and E. M. Landis are the inventors of the algorithm
// bearing their initials, AVL.  Their algorithm for height-balanced binary
// search trees is the basis of the code in this package.  Some modification
// is made to the balancing operations, and features are added for practical
// purposes.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		avl.h
//
// purpose	Define resources for the AVL binary tree package
//-----------------------------------------------------------------------------
#ifndef __AVL_H__
#define __AVL_H__

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

//-----------------------------------------------------------------------------
// configuration
//-----------------------------------------------------------------------------
#ifndef __PROTOTYPE_BEGIN__
#define __PROTOTYPE_BEGIN__
#endif
#ifndef __PROTOTYPE_END__
#define __PROTOTYPE_END__
#endif

#ifndef AVL_COUNT_ENABLE
#define AVL_COUNT_ENABLE                 0
#endif

//-----------------------------------------------------------------------------
// structs
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// struct	avl_link
//
// purpose	Contain the links and other information that implements one
//		binary tree structuring instance for a node structure that
//		can be defined by the calling program.
//-----------------------------------------------------------------------------
struct avl_link {
    struct avl_link *		lo	;
    struct avl_link *		hi	;
    struct avl_link *		pa	;
    int				depth	;
};
typedef struct avl_link		avl_link	;

//-----------------------------------------------------------------------------
// struct	avl_root
//
// purpose	This struct is the root node for a balanced binary tree.
//		The root node does not have a key.  It is parented from the
//		middle node located at the apex of the tree (which would be
//		the root, otherwise) and is effectively positioned as if it
//		was keyed lower in value than the lowest possible member of
//		the tree.
//
//		A reference to a binary tree is a pointer to this struct.
//-----------------------------------------------------------------------------
struct avl_root {

//-----------------------------------------------------------------------------
// field	link
//
// The control struct for the binary tree is also the root node of the tree
// structure.  This field contains the links necessary to allow this.  The
// root node is always positioned as the parent of the apex of the regular
// nodes, and the apex is linked via the hi pointer of the root.  This makes
// the root positioned as always lower in key value than the lowest possible
// value a key could really have.
//-----------------------------------------------------------------------------
    avl_link		link		;// link member for this pseudo-node

//-----------------------------------------------------------------------------
// field	select
//
// Whenever a node is found or inserted in the binary tree, it becomes the
// currently selected node.  This field points to the links for the currently
// selected node.  Certain operations can refer to the currently selected
// node when no node is explicitly provided.
//-----------------------------------------------------------------------------
    avl_link *		select		;// currently selected node

//-----------------------------------------------------------------------------
// field	offset
//
// The calling program can work with node structs in which the links are not
// actually located at the beginning of the struct.  Internally, AVL always
// works with pointers that point exactly to the links sub-struct.  The offset
// translates caller node pointers to internal link pointers by addition, and
// translates internal link pointers to caller node pointers by subtraction.
//-----------------------------------------------------------------------------
    size_t		offset		;// offset for caller pointers

//-----------------------------------------------------------------------------
// field	compare_f
//
// The calling program provides a function which compares the keys in two
// key link structures as defined by the calling program.  This field keeps
// the pointer to that function.
//-----------------------------------------------------------------------------
    int			( * compare_f )( avl_link *, avl_link *, struct avl_root * );

#if AVL_COUNT_ENABLE
//-----------------------------------------------------------------------------
// field	count_nodes
//
// Keep a count of the number of nodes in a tree, for consistency check.
//-----------------------------------------------------------------------------
    unsigned long	count_nodes	;// count of nodes in tree
#endif /* AVL_COUNT_ENABLE */

//-----------------------------------------------------------------------------
// field	err_num
//
// This field records an error code identifying the last error to occur with
// this binary tree.
//-----------------------------------------------------------------------------
    int			err_num		;// last error in this tree
};
typedef struct avl_root		avl_root		;

//-----------------------------------------------------------------------------
// Define the official types that calling programs should use to reference
// a binary tree, or create an instance of a binary tree root node.
//-----------------------------------------------------------------------------
typedef struct avl_root		AVL		;
typedef struct avl_root *	AVLP		;
typedef struct avl_root *	AVL_P		;

//-----------------------------------------------------------------------------
// error symbols
//-----------------------------------------------------------------------------
#define AVL_OK				0

#define AVL_ERR_BAD_ROOT		-1
#define AVL_ERR_BAD_NODE		-2
#define AVL_ERR_ROOT_NULL		-3
#define AVL_ERR_NODE_NULL		-4
#define AVL_ERR_NO_PARENT		-5

#define AVL_ERR_CORRUPT_PA_NOBACK	0x0001
#define AVL_ERR_CORRUPT_PA_NULL		0x0002
#define AVL_ERR_CORRUPT_LOHI_CROSSED	0x0004
#define AVL_ERR_CORRUPT_LO_NOBACK	0x0010
#define AVL_ERR_CORRUPT_HI_NOBACK	0x0040
#define AVL_ERR_CORRUPT_DEPTH		0x0100

//-----------------------------------------------------------------------------
// count macros
//-----------------------------------------------------------------------------
#if AVL_COUNT_ENABLE

#define avl_count_dec_nodes(r)		(--((r)->count_nodes))
#define avl_count_inc_nodes(r)		(++((r)->count_nodes))
#define avl_count_get_nodes(r)		((r)->count_nodes)
#define avl_count_set_nodes(r,n)	(((r)->count_nodes)=(n))

#else /* AVL_COUNT_ENABLE */

#define avl_count_dec_nodes(r)
#define avl_count_inc_nodes(r)
#define avl_count_get_nodes(r)	0
#define avl_count_set_nodes(r,n)

#endif /* AVL_COUNT_ENABLE */


//-----------------------------------------------------------------------------
// macros
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// macro	avl_find_exact
//
// purpose	Find the first exact link in a binary tree without any
//		influence toward lower or higher positions in multiple
//		instances of the same key.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//
// returns	(void *) pointer to node with any exact matching key.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
// macro	avl_find_exact_lo
//
// purpose	Find the exact link in a binary tree that is in the lowest
//		position among multiple instances of the same key.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//
// returns	(void *) pointer to node with the lowest exact matching key.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
// macro	avl_find_exact_hi
//
// purpose	Find the exact link in a binary tree that is in the highest
//		position among multiple instances of the same key.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//
// returns	(void *) pointer to node with the highest exact matching key.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
// macro	avl_find_lo_or_near
//
// purpose	Find the exact link in a binary tree that is in the lowest
//		position among multiple instances of the same key, or if no
//		exact key can be found, the nearest key lower than given.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//
// returns	(void *) pointer to node with the lowest exact matching key.
//		(void *) pointer to node with nearest lower key if no exact.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
// macro	avl_find_hi_or_near
//
// purpose	Find the exact link in a binary tree that is in the highest
//		position among multiple instances of the same key, or if no
//		exact key can be found, the nearest key higher than given.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//
// returns	(void *) pointer to node with the highest exact matching key.
//		(void *) pointer to node with nearest higher key if no exact.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
#define avl_find_lo_or_near(t,k)	avl_find((t),(k),-2)
#define avl_find_exact_lo(t,k)		avl_find((t),(k),-1)
#define avl_find_exact(t,k)		avl_find((t),(k),0)
#define avl_find_exact_hi(t,k)		avl_find((t),(k),1)
#define avl_find_hi_or_near(t,k)	avl_find((t),(k),2)

//-----------------------------------------------------------------------------
// macro	avl_loop_forward
//
// purpose	Code a loop that steps through the nodes of a binary tree in
//		in the forward direction from first to last.
//
// arguments	1 (AVL *) pointer to the binary tree (root node)
//		2 (node *) name of variable to have the node pointer in the loop
//-----------------------------------------------------------------------------
#define avl_loop_forward(t,p) for(avl_unselect((t)),(p)=(void *)(NULL);((p)=(void *)avl_forward((t),(void *)(p)));)

//-----------------------------------------------------------------------------
// macro	avl_loop_reverse
//
// purpose	Code a loop that steps through the nodes of a binary tree in
//		in the reverse direction from last to first.
//
// arguments	1 (AVL *) pointer to the binary tree (root node)
//		2 (node *) name of variable to have the node pointer in the loop
//-----------------------------------------------------------------------------
#define avl_loop_reverse(t,p) for(avl_unselect((t)),(p)=(void *)(NULL);((p)=(void *)avl_reverse((t),(void *)(p)));)

//-----------------------------------------------------------------------------
// macro	avl_init
//
// purpose	Initialize a binary tree for use with a specific node type
//		and a specific field for the tree links.
//
// arguments	1 (AVL *) pointer to the binary tree
//		2 ... struct type caller will use as a node for this tree
//		3 ... links field of node struct this tree will index with
//		4 (int (* fun)(void *,void *)) function to do key compares
//
// returns	(AVL *) pointer to same root struct now initialized
//		(AVL *) NULL if there is an error
//-----------------------------------------------------------------------------
#define avl_init(b,t,f,c) avl_initialize((b),((size_t)(&(((t*)0)->f))),(c))

//-----------------------------------------------------------------------------
// macro	avl_is_empty
//
// purpose	Yield true if the tree is empty, or false if it is not empty.
//
// arguments	1 (AVL *) pointer to binary tree
//
// returns	(int) true if tree is empty
//		(int) false if tree is not empty
//-----------------------------------------------------------------------------
#define avl_is_empty(t) (!((t)->link.hi))

//-----------------------------------------------------------------------------
// macro	avl_is_not_empty
//
// purpose	Yield true if the tree is not empty, or false if it is empty.
//
// arguments	1 (AVL *) pointer to binary tree
//
// returns	(int) true if tree is empty
//		(int) false if tree is not empty
//-----------------------------------------------------------------------------
#define avl_is_not_empty(t) (!(!((t)->link.hi)))

//-----------------------------------------------------------------------------
// macro	avl_insert
//
// purpose	Call avl_put to do an insert only, without duplicate.
//-----------------------------------------------------------------------------
// macro	avl_insert_dup
//
// purpose	Call avl_put to do an insert with duplication allowed.
//-----------------------------------------------------------------------------
// macro	avl_replace
//
// purpose	Call avl_put to do an insert or replacement.
//-----------------------------------------------------------------------------
#define avl_insert(t,n)		avl_put((t),(n),NULL,0)
#define avl_insert_dup(t,n)	avl_put((t),(n),NULL,1)
#define avl_replace(t,n,s)	avl_put((t),(n),(s),0)

//-----------------------------------------------------------------------------
// functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// function	avl_abort
//
// purpose	Abort program or process execution after outputting diagnostic
//		information about the cause for the problem.
//
// arguments	1 (AVL *) pointer to binary tree where problem was found
//		2 (avl_link *) pointer to link around which problem was found
//		3 (const char *) message format string
//		4 ... variable arguments for message
//
// returns	(int) 1 only if environment variable AVL_ABORT_BYPASS is set
//		otherwise does not return
//-----------------------------------------------------------------------------
int
avl_abort (
    AVL *		arg_tree
    ,
    avl_link *		arg_link
    ,
    const char *	arg_message
    ,
    ...
    )
;
//-----------------------------------------------------------------------------
// function	avl_balance
//
// purpose	Balance the binary tree branch containing the specified node.
//
// arguments	1 (AVL *) pointer to root of the binary tree
//		2 (avl_link *) pointer to node to do balancing from
//
// returns	(int) number of shuffles performed
//-----------------------------------------------------------------------------
int
avl_balance(
    AVL *		arg_tree
    ,
    avl_link *		arg_link
    )
;
//-----------------------------------------------------------------------------
// function	avl_check_link
//
// purpose	Perform a diagnostic check on the correct structure of a
//		binary tree around a given link.  This function	uses link
//		pointers instead of node pointers, and so is intended for
//		internal use only.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (avl_link *) pointer to link in node to check
//
// returns	(int) 0 for no error
//		(int) or-ed flags for detected errors
//-----------------------------------------------------------------------------
int
avl_check_link (
    AVL *	arg_tree
    ,
    avl_link *	arg_link
    )
;
//-----------------------------------------------------------------------------
// function	avl_check_node
//
// purpose	Perform a diagnostic check on the correct structure of a
//		binary tree around a given node.  If there is an error,
//		abort program execution unless the environment variable
//		AVL_ERROR_CONTINUE is defined non-zero or empty.
//
// arguments	1 (AVL *) pointer to binary tree to check
//		2 (void *) pointer to node (not link)
//
// returns	(int) 0 for no error
//		(int) or-ed flags for detected errors
//-----------------------------------------------------------------------------
int
avl_check_node (
    AVL *		arg_tree
    ,
    void *		arg_node
    )
;
//-----------------------------------------------------------------------------
// function	avl_check_tree
//
// purpose	Perform a diagnostic check on the correct structure of a
//		whole binary tree.  If there is an error, abort program
//		execution unless the environment variable AVL_ERROR_CONTINUE
//		is defined non-zero or empty.
//
// arguments	1 (AVL *) pointer to binary tree to check
//
// returns	(int) 0 for no error
//		(int) or-ed flags for detected errors
//-----------------------------------------------------------------------------
int
avl_check_tree (
    AVL *		arg_tree
    )
;
//-----------------------------------------------------------------------------
// function	avl_delete
//
// purpose	Delete a specified node or the current node from a given
//		binary tree.
//
// arguments	1 (AVL *) pointer to root of the binary tree
//		2 (void *) pointer to node to delete, or NULL to delete the
//			currently selected node
//
// returns	(void *) pointer to deleted node.
//		(void *) NULL if no node deleted.
//
// note		The caller is responsible for returning the memory space
//		occupied by the deleted node if it was dynamically allocated.
//-----------------------------------------------------------------------------
void *
avl_delete (
    AVL *		arg_tree
    ,
    void *		arg_node
    )
;
//-----------------------------------------------------------------------------
// function	avl_empty
//
// purpose	Delete all nodes from the specified binary tree in a faster
//		way than would be done if each node were deleted individually.
//
//		Since the time to delete one node is almost O(log N) then the
//		time to delete all nodes one at a time is more than O(N) (but
//		not quite O(N log N) since N is diminishing), this function
//		is preferred as a faster way to delete all nodes in bulk since
//		its execution time is proportional to O(N).
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void (*fun)(void *)) optional pointer to function to call
//
// returns	(long) number of nodes deleted
//		(long) < 0 : error
//
// callback	The function specified in argument 2 is called with:
//   arguments	1 (void *) pointer to node
//   returns	(void) ignored
//
// note		Although the function pointer in argument 2 is optional, it
//		is essential to specify a function if the nodes need to be
//		returned to free memory.  The avl_empty function will not free
//		the nodes otherwise.  Just specify free() as the function to
//		call to simply free all the nodes as they are deleted, if no
//		other special handling is needed for each deleted node.  Or
//		specify your own function that uses the callback arguments.
//-----------------------------------------------------------------------------
long
avl_empty (
    AVL *		arg_tree
    ,
    void		(*(arg_fun))(void *)
    )
;
//-----------------------------------------------------------------------------
// function	avl_find
//
// purpose	Find a link in a binary tree based on a key in a given link
//		structure.
//
//		If the influential direction is none, return the first exact
//		match encountered, or return none.
//
//		If the influential direction is low, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		low direction.
//
//		If the influential direction is high, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		high direction.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to argument key link
//		3 (int) direction and near option, -2, -1, 0, 1, 2
//			-2 = choose lowest exact, or nearest lower
//			-1 = choose lowest exact only
//			 0 = choose fastest exact only
//			 1 = choose highest exact only
//			 2 = choose highest exact, or nearest higher
//
// returns	(void *) pointer to node with matching key.
//		(void *) NULL if node not found
//-----------------------------------------------------------------------------
void *
avl_find (
    AVL *	arg_tree
    ,
    void *	arg_key
    ,
    int		arg_dir_near
    )
;
//-----------------------------------------------------------------------------
// function	avl_first
//
// purpose	Find the first node with the lowest key value from the given
//		binary tree.
//
// arguments	1 (AVL *) pointer to binary tree
//
// returns	(void *) pointer to the first node
//		(void *) NULL if the binary tree is empty
//-----------------------------------------------------------------------------
void *
avl_first (
    AVL *		arg_tree
    )
;
//-----------------------------------------------------------------------------
// function	avl_last
//
// purpose	Find the last node with the highest key value from the given
//		binary tree.
//
// arguments	1 (AVL *) pointer to binary tree
//
// returns	(void *) pointer to the last node
//		(void *) NULL if the binary tree is empty
//-----------------------------------------------------------------------------
void *
avl_last (
    AVL *		arg_tree
    )
;
//-----------------------------------------------------------------------------
// function	avl_forward
//
// purpose	Find the node with the next higher key from the specified
//		binary tree.  The type of key is irrelevant.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) optional pointer to node to start from
//
// returns	(void *) pointer to the next higher node
//		(void *) NULL if there is no higher node
//-----------------------------------------------------------------------------
void *
avl_forward (
    AVL *	arg_tree
    ,
    void *	arg_node
    )
;
//-----------------------------------------------------------------------------
// function	avl_reverse
//
// purpose	Find the node with the previous lower key from the specified
//		binary tree.  The type of key is irrelevant.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) optional pointer to node to start from
//
// returns	(void *) pointer to the next lower node
//		(void *) NULL if there is no lower node
//-----------------------------------------------------------------------------
void *
avl_reverse (
    AVL *	arg_tree
    ,
    void *	arg_node
    )
;
//-----------------------------------------------------------------------------
// function	avl_initialize
//
// purpose	Initialize the root struct of a binary tree to the initial
//		empty state, ready for operations, regardless of the original
//		contents.  The original contents will be lost, and if any
//		pointers are the only pointers to allocated memory, then that
//		memory is also lost.
//
// arguments	1 (AVL *) pointer to root struct to be initialized
//		2 (size_t) offset from node struct to links struct
//		3 (int (* fun)(void *,void *)) function to do key compares
//
// returns	(AVL *) pointer to same root struct now initialized
//		(AVL *) NULL if there is an error
//-----------------------------------------------------------------------------
AVL *
avl_initialize(
    AVL *		arg_tree
    ,
    size_t		arg_offset
    ,
    int			(* arg_compare_f)(avl_link *,avl_link *,AVL *)
)
;
//-----------------------------------------------------------------------------
// function	avl_iterate
//
// purpose	Execute a given function once for each node in the given
//		binary tree.  Iteration is done in forward order.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void (*fun)(void *)) pointer to function to execute
//
// returns	(long) sum of return values or error code.
//
// callback	The function specified in argument 2 is called with:
//   arguments	1 (void *) pointer to node
//   returns	(void) ignored
//
// warning	Do NOT delete the current node from the tree during the
//		function being called with it.
//-----------------------------------------------------------------------------
long
avl_iterate(
    AVL *		arg_tree
    ,
    int			( (* arg_fun)(void *) )
    )
;
//-----------------------------------------------------------------------------
// function	avl_print_link
//
// purpose	Print a binary tree node link.
//
// arguments	1 (FILE *) output file
//		2 (int) width limit to truncate to
//		3 (avl_link *) link to print
//
// returns	(int) width actually used
//-----------------------------------------------------------------------------
int
avl_print_link (
    FILE *		arg_file
    ,
    int			arg_limit
    ,
    avl_link *		arg_link
    )
;
//-----------------------------------------------------------------------------
// function	avl_print_tree
//
// purpose	Print all the links in a branch of a binary tree.
//		Call self for sub-branches.
//
// arguments	1 (FILE *) file to print to
//		2 (AVL *) pointer to binary tree
//		3 (void *) pointer to node at apex of branch to print
//		4 (const char *) prefix to print in front of each node
//		5 (int) option to print link pointers
//		6 (const char * (*fun)(void *)) function to get suffix to print
//		7 (int) maximum width to truncate line
//		8 (const char *) string to append after line truncation
//		9 (int) maximum rows to truncate screen
//
// returns	(int) number of rows printed
//
// note		Recursion is used here so that this function is more robust
//		with corrupt trees, making it more useful for debugging.
//-----------------------------------------------------------------------------
int
avl_print_tree (
    FILE *		arg_file
    ,
    AVL *		arg_tree
    ,
    void *		arg_node
    ,
    const char *	arg_prefix
    ,
    int			arg_ptrs
    ,
    const char *	(*arg_fun)(void *)
    ,
    int			arg_width
    ,
    const char *	arg_append
    ,
    int			arg_rows
    )
;
//-----------------------------------------------------------------------------
// function	avl_print_link_detail
//
// purpose	Print information about a node in detail, including nodes
//		it points to.
//
// arguments	1 (FILE *) open file to print to
//		2 (AVL *) tree pointer for reference
//		3 (avl_link *) pointer to link to print
//
// returns	(int) 0
//-----------------------------------------------------------------------------
int
avl_print_link_detail(
    FILE *		arg_file
    ,
    AVL *		arg_tree
    ,
    avl_link *		arg_link
    )
;
//-----------------------------------------------------------------------------
// function	avl_put
//
// purpose	Put a node into the given binary tree at the location
//		indicated by the key link within the node.
//
// arguments	1 (AVL *) pointer to binary tree
//		2 (void *) pointer to node
//		3 (void * *) where to store pointer to replaced node
//		4 (int) option: allow duplicate insert
//
// note		If argument 3 is non-NULL, then replacement is allowed.
//		If argument 3 is NULL, then replacement is not allowed.
//
// returns	(int) -1: node not replaced or inserted
//		(int)  0: node replaces an existing node
//		(int)  1: node inserted as new
//		(int)  2: node inserted as duplicate
//
// note		If the allow duplicate insert option is true, then the node
//		is inserted regardless of whether one already exists with
//		that key or not.  The order duplicate nodes are retrieved,
//		and which node will match a find, is not defined.
//-----------------------------------------------------------------------------
// macro	avl_insert
//
// purpose	Call avl_put to do an insert only, without duplicate.
//-----------------------------------------------------------------------------
// macro	avl_insert_dup
//
// purpose	Call avl_put to do an insert with duplication allowed.
//-----------------------------------------------------------------------------
// macro	avl_replace
//
// purpose	Call avl_put to do an insert or replacement.
//-----------------------------------------------------------------------------
int
avl_put (
    AVL *	arg_tree
    ,
    void *	arg_node
    ,
    void * *	arg_rep_node
    ,
    int		arg_dup_ok
    )
;
//-----------------------------------------------------------------------------
// function	avl_unselect
//
// purpose	Clear the selection of the current node so that there is
//		no node currently selected.  This causes, for example, the
//		avl_forward() function to get the first node the next time
//		it is called for this tree.
//
// arguments	1 (AVL *) pointer to root of the binary tree
//
// returns	(void *) pointer to the previously selected node
//		(void *) NULL if no node was previously selected
//-----------------------------------------------------------------------------
void *
avl_unselect (
    AVL *		arg_tree
    )
;

//-----------------------------------------------------------------------------
// end		avl.h
//-----------------------------------------------------------------------------
#endif /* __AVL_H__ */
