//-----------------------------------------------------------------------------
// Copyright  2000 - Philip Howard - All rights reserved
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//-----------------------------------------------------------------------------
// package	avlmap
// homepage	http://phil.ipal.org/freeware/
//-----------------------------------------------------------------------------
// author	Philip Howard
// email	phil@ipal.org
// homepage	http://phil.ipal.org/
//-----------------------------------------------------------------------------
// This file is best viewed using a fixed spaced font such as Courier
// and in a display at least 112 columns wide.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// file		map_scalar.h
//
// purpose	Declare common functions for scalar data type implementation.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// The following must be defined by the file that includes this one:
//
// map_type		Typedef this to the C type for the scalar type to be
//			used for the key or data, e.g. unsigned long.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// function	map_xx_new
//
// purpose	Create a new mapping with the specific scalar key type.
//
// arguments	1 (int *)() optional alternate comparison function
//
// returns	(MAP) handle for the new mapping.
//		(MAP) NULL = error creating new mapping
//-----------------------------------------------------------------------------
MAP
map_xx_new ();

//-----------------------------------------------------------------------------
// function	map_xx_find
//
// purpose	Find a node with a matching key, or possibly a nearby key.
//
//		If the influential direction is none, return the first exact
//		match encountered, or return none.
//
//		If the influential direction is low, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		low direction.
//
//		If the influential direction is high, return the lowest exact
//		match encountered, or if there is no exact match: if the
//		near option is specified, choose the nearest node in the
//		high direction.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key
//		3 (int) direction and near option, -2, -1, 0, 1, 2
//			-2 = choose lowest exact, or nearest lower
//			-1 = choose lowest exact only
//			 0 = choose first exact only
//			 1 = choose highest exact only
//			 2 = choose highest exact, or nearest higher
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find (
    MAP			arg_map
    ,
    map_type		arg_key
    ,
    int			arg_dir_near
);

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_lo
//
// purpose	Find a node with an exactly matching key of the specific
//		scalar key type for this mapping.  If there is more than
//		one node of the same key, select the one in the lowest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_exact_lo (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_find_lo_or_near
//
// purpose	Find a node with an exactly matching key, or if no exactly
//		matching key present, a node with nearest lower value, of
//		the specific scalar key type for this mapping.  If there is
//		more than one node of the same key, select the one in the
//		lowest position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_lo_or_near (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_find_exact_hi
//
// purpose	Find a node with an exactly matching key of the specific
//		scalar key type for this mapping.  If there is more than
//		one node of the same key, select the one in the highest
//		position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_exact_hi (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_find_hi_or_near
//
// purpose	Find a node with an exactly matching key, or if no exactly
//		matching key present, a node with nearest higher value, of
//		the specific scalar key type for this mapping.  If there is
//		more than one node of the same key, select the one in the
//		highest position.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = no node found
//		(int)  0 = found a node with no data
//		(int) data type of found node
//-----------------------------------------------------------------------------
int
map_xx_find_hi_or_near (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_insert
//
// purpose	Insert a new member into a mapping with the specified key and
//		an initial undefined data value, unless one already exists
//		with that exact key, in which case it is made active.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int)  0 = found or inserted a member with no data
//		(int) >0 = data type of found node
//-----------------------------------------------------------------------------
int
map_xx_insert (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_insert_dup
//
// purpose	Insert a new node into a mapping with the specified key and
//		an initial undefined data value.  If a node already exists,
//		insert a duplicate.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) scalar key type
//
// returns	(int) -2 = error, no map, bad type
//		(int) -1 = not inserted
//		(int)  1 = new unique node inserted
//		(int)  2 = duplicate node inserted
//-----------------------------------------------------------------------------
int
map_xx_insert_dup (
    MAP			arg_map
    ,
    map_type		arg_key
);

//-----------------------------------------------------------------------------
// function	map_xx_key
//
// purpose	Get the key value of the current node in the mapping.
//
// arguments	1 (MAP) mapping handle
//
// return	(map_type) key value
//		(map_type) 0 cast to the key type of the tree
//-----------------------------------------------------------------------------
map_type
map_xx_key (
    MAP			arg_map
);

//-----------------------------------------------------------------------------
// function	map_xx_store
//
// purpose	Store the given data into the current node in the mapping.
//
// arguments	1 (MAP) mapping handle
//		2 (map_type) data value to store
//
// return	(int) -2 = error, no map, no node
//		(int)  0 = success, data stored
//-----------------------------------------------------------------------------
int
map_xx_store (
    MAP			arg_map
    ,
    map_type		arg_data
);

//-----------------------------------------------------------------------------
// function	map_xx_fetch
//
// purpose	Fetch the data from the current node in the mapping and return
//		directly as a value.  There is no means to indicate an error.
//
// arguments	1 (MAP) mapping handle
//
// return	(map_type) data value returned
//		(map_type) 0 cast to the data type if error
//-----------------------------------------------------------------------------
map_type
map_xx_fetch (
    MAP			arg_map
);
